package de.ifgi.main;

import java.util.ArrayList;
import java.util.List;

import de.ifgi.gui.Parameter;
import de.ifgi.operations.KoCriteria;
import de.ifgi.operations.SimilaritySelector;
import de.ifgi.operations.SpatialChecker;
import de.ifgi.operations.Tiffloader;
import de.ifgi.proxy.GeotiffProxy;

public class ImageManager {
	private static Tiffloader tiffLoader = Tiffloader.getInstance();
	private static KoCriteria KO_Checker = KoCriteria.getInstance();
	private static SpatialChecker spatialChecker = SpatialChecker.getInstance();
	private static SimilaritySelector similaritySelector = SimilaritySelector.getInstance();
	
    private static ImageManager instance;

    private ImageManager() {}

    public synchronized static ImageManager getInstance() {
        if (instance == null) {
            instance = new ImageManager();
        }
        return instance;
    }

	
	public void selectImages(Parameter parameter){
		String[] inputFiles = parameter.getFiles();
		double selectionQuota = parameter.getSelectionQuota();
		double filterQuota = parameter.getFilterQuota();
		List<GeotiffProxy> checkedImages = loadGeotiffs(inputFiles,filterQuota);
		checkedImages = checkKOcriteria(checkedImages);
		List<List<GeotiffProxy>> spatialSets = checkSpatially(checkedImages);
		List<GeotiffProxy> finalSelection = new ArrayList<GeotiffProxy>();
		for(List<GeotiffProxy> spatialSet : spatialSets ){
			finalSelection.addAll(selectFromSet(selectionQuota,spatialSet));
		}
		writeGeotiffs(finalSelection);
		
	}
	
	public void selectImages(double selectionQuota, List<GeotiffProxy> checkedImages){
		
		checkedImages = checkKOcriteria(checkedImages);
		List<List<GeotiffProxy>> spatialSets = checkSpatially(checkedImages);
		List<GeotiffProxy> finalSelection = new ArrayList<GeotiffProxy>();
		for(List<GeotiffProxy> spatialSet : spatialSets ){
			finalSelection.addAll(selectFromSet(selectionQuota,spatialSet));
		}
		writeGeotiffs(finalSelection);
		
	}
	
	private List<GeotiffProxy> loadGeotiffs(String[] files, double sampleQuota){
		return tiffLoader.loadGeotiffs(sampleQuota,files);
	}
	
	private List<GeotiffProxy> checkKOcriteria(List<GeotiffProxy> checkedImages){
		return KO_Checker.checkKOcriteria(checkedImages);
	}
	
	private List<List<GeotiffProxy>> checkSpatially(List<GeotiffProxy> checkedImages){
		return spatialChecker.checkSpatially(checkedImages);
	}

	private List<GeotiffProxy> selectFromSet(double selectionQuota, List<GeotiffProxy> spatialSet){
		return similaritySelector.selectFromSet(selectionQuota, spatialSet);
	}

	private void writeGeotiffs(List<GeotiffProxy> finalSelection) {
		tiffLoader.writeGeotiffs(finalSelection);
	}


	
}
