package de.ifgi.operations;

import java.awt.FileDialog;
import java.awt.Frame;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.rosuda.JRI.REXP;
import org.rosuda.JRI.RMainLoopCallbacks;
import org.rosuda.JRI.Rengine;

import de.ifgi.proxy.GeotiffProxy;
import de.ifgi.proxy.SampleElement;



public class SimilaritySelector {

    private static SimilaritySelector instance;

    private SimilaritySelector() {}

    public synchronized static SimilaritySelector getInstance() {
        if (instance == null) {
            instance = new SimilaritySelector();
        }
        return instance;
    }
	
	public Rengine re=new Rengine(null, false, new TextConsole());
	
	public List<GeotiffProxy> selectFromSet(double selectionQuota,  List<GeotiffProxy> spatialSet) {
		//number of images
		int imNo = spatialSet.size();
		//number of channels
		int chNo = spatialSet.get(0).getSample()[0].getNumberOfChannels();
		double[][] finalDistanceMatrixMahalanobis = new double[imNo][imNo];
		for(int i = 0; i<chNo; i++){
			//estimate a_coefficient for deriving the covariance matrix
	        double a_coeff = estimate_A_coeff(i,spatialSet);
	        
	        //calculate variances
	        //create an array of randomly chosen samples to use it in R
			int[] sampleArray = createAllSamplesArray(i, spatialSet);
						
			//calculate the variances with R
			long r_sampleArray = re.rniPutIntArray(sampleArray);
			re.rniAssign("allsamples", r_sampleArray, 0);
	        REXP r_varVector = re.eval("varVector<-diag(cov(t(matrix(allsamples, ncol="+spatialSet.size()+"))))");
	        
	        double[] varVector = r_varVector.asDoubleArray();
	        
	        //use the a -coefficient and the variances to derive the final covariance matrix S
	        double[][] S = calculateFinalCovMatrix(spatialSet.get(0).getSample(), a_coeff, varVector);
	        
	        //transform S to R-Matrix
	        double[] temp = doubleMatrixInRMatrix(S);
	        long r_temp = re.rniPutDoubleArray(temp);
			re.rniAssign("temp", r_temp, 0);
	        re.eval("S<-matrix(temp, ncol="+S.length+")");
	        
	        //invert S
	        REXP r_S_inv = re.eval("S_inv<-solve(S)");
	        
	        //create mahalanobis distance matrix
	        double[][] tmpDistanceMatrixMahalanobis = createDistanceMatrixMahalanobis(i, spatialSet, r_S_inv);
	        finalDistanceMatrixMahalanobis = addMatrices(tmpDistanceMatrixMahalanobis,finalDistanceMatrixMahalanobis);
		}
        
        //transform distanceMatrixMahalanobis to R-Matrix
        double[] temp = doubleMatrixInRMatrix(finalDistanceMatrixMahalanobis);
        long r_temp = re.rniPutDoubleArray(temp);
		re.rniAssign("temp", r_temp, 0);
        re.eval("dissimMatrix<-matrix(temp, ncol="+finalDistanceMatrixMahalanobis.length+")");
        
        //perform a multidimensional scaling to find out dissimilarities
        REXP r_mdsResult = re.eval("mdsresult<-cmdscale(dissimMatrix,k=1)");
        
        //set MDS Values and order the set of geotiffs by its mdsvalue
        double[] mds_result = r_mdsResult.asDoubleArray();
        setMDSValues(mds_result,spatialSet);
        
        List<GeotiffProxy> orderedSpatialSet = orderGeotiffByMDS(spatialSet);
        
        //calculate the number of pictures to be selected
        int outputNumber = new Double(selectionQuota*orderedSpatialSet.size()).intValue();
         
        List<GeotiffProxy> outputGeotiffs = getOutputGeotiffs(outputNumber, orderedSpatialSet);
        
        return outputGeotiffs;
	}

	private double[][] addMatrices(double[][] m1, double[][] m2) {
		double[][] result = new double[m1.length][m1.length];
		for(int i = 0; i<m1.length; i++){
			for(int j = 0; j<m1.length; j++){
				result[i][j] = m1[i][j]+m2[i][j];
			}
		}
		return result;
	}

	private List<GeotiffProxy> getOutputGeotiffs(int outputNumber,
			List<GeotiffProxy> orderedSpatialSet) {
		double inputNumber = new Double(orderedSpatialSet.size());
		List<GeotiffProxy> outputGeotiffs = new ArrayList<GeotiffProxy>(outputNumber);
		
		double stepwidth = (inputNumber-1)/(new Double(outputNumber-1)); 
		double position = 0.0;
		for(int i = 0; i<outputNumber; i++){
			outputGeotiffs.add(orderedSpatialSet.get(new Double(position).intValue()));
			position = position + stepwidth;
		}
		
		return outputGeotiffs;
	}

	private List<GeotiffProxy> orderGeotiffByMDS(List<GeotiffProxy> spatialSet) {
		List<GeotiffProxy> orderedList = new ArrayList<GeotiffProxy>(spatialSet.size());
		
		while(!spatialSet.isEmpty()){
			GeotiffProxy smallest = spatialSet.get(0);
			
			for(GeotiffProxy geotiff : spatialSet){
				if(geotiff.getMDSValue()< smallest.getMDSValue()){
					smallest = geotiff;
				}
			}
			orderedList.add(smallest);
			spatialSet.remove(smallest);
		}
		
		return orderedList;
	}

	private void setMDSValues(double[] mdsResult, List<GeotiffProxy> spatialSet) {
		int i = 0;
		for(GeotiffProxy geotiff : spatialSet){
			geotiff.setMDSValue(mdsResult[i]);
			i++;
		}
	}

	private double estimate_A_coeff(int chNo, List<GeotiffProxy> spatialSet) {
		int n = 30;
		Random r = new Random();
		int q = spatialSet.get(0).getSample().length;
		
		//randomly chosen positions to pick the sample elements used to estimate the regression afterwards
		int[] sampleMask = new int[n];
		for(int i = 0; i<n; i++){
			sampleMask[i] = r.nextInt(q);
		}
		
		//create an array of randomly chosen samples to use it in R
		int[] sampleArray = createSamplesArray(chNo, spatialSet, sampleMask);
		
		
		//calculate the correlation matrix with R, we need that as regression values for cov(h)
		long r_sampleArray = re.rniPutIntArray(sampleArray);
		re.rniAssign("allsamples", r_sampleArray, 0);
        REXP r_corrMatrix = re.eval("corMatrix<-cor(t(matrix(allsamples, ncol="+spatialSet.size()+")))");
        
        //create a vector of pairs of distance and corrValue between two pixels
        double[] distanceVector = createDistanceVector(spatialSet, sampleMask);
        double[] corrVector = createCovarianceVector(r_corrMatrix);
        
        //estimate e^(-h/a)
        long r_distanceVector = re.rniPutDoubleArray(distanceVector);
		re.rniAssign("dist", r_distanceVector, 0);
		
		long r_covVector = re.rniPutDoubleArray(corrVector);
		re.rniAssign("corrvec", r_covVector, 0);
		REXP r_coeff;
		double a = 1;
		//if the starting value is chosen wrongly, R crashes. Therefore it has to be varied!
		do{
			re.eval("reg<-nls(corrvec ~ exp(-dist/a), start = list(a = "+a+"))");
	        r_coeff = re.eval("coef(reg)");
	        a = a + 1;	
		} while (r_coeff == null);
		
		return r_coeff.asDouble();
	}

	private double[][] createDistanceMatrixMahalanobis(
			int chNo, List<GeotiffProxy> spatialSet, REXP r_S_Inv) {

		
		double[][] distanceMatrix = new double[spatialSet.size()][spatialSet.size()];
		int n = spatialSet.get(0).getSample().length;
		double[][] S_inv = new double[n][n];
		// sometimes there can occur computational problems. in that case, use euclidean distance. (S = I)
		if (r_S_Inv == null){
			for(int i = 0; i < n; i++){
				S_inv[i][i]=1;
			}
		}
		else{
			S_inv = r_S_Inv.asDoubleMatrix();
		}
		for(int i = 0; i<distanceMatrix.length; i++){
			for(int j = i; j<distanceMatrix.length; j++){
				double tempDistance = calculateDistanceMahalanobis(chNo,spatialSet.get(i),spatialSet.get(j),S_inv);
				distanceMatrix[i][j] = tempDistance;
				distanceMatrix[j][i] = tempDistance;
			}
		}
		return distanceMatrix;
	}

	private double calculateDistanceMahalanobis(int chNo, GeotiffProxy g1,
			GeotiffProxy g2, double[][] sInv) {
		double distance=0;

		if(g1.equals(g2)){
			return distance;
		}
		else{
			//difference vector d between two images
			double[] d = new double[g1.getSample().length];
			for(int i = 0; i<d.length; i++){
				d[i] = g1.getSample()[i].getValue(chNo)-g2.getSample()[i].getValue(chNo);
			}
			
			//multiply S inverse with d
			double[] S_inv_mult_d = new double[sInv.length];
			for(int i = 0; i< S_inv_mult_d.length; i++){
				S_inv_mult_d[i] = 0;
				for(int j = 0; j<d.length; j++){
					S_inv_mult_d[i] = S_inv_mult_d[i]+ sInv[i][j]*	d[j];				
				}
			}
			
			//multiply d with S^(-1)*d
			for(int i = 0; i<d.length; i++){
				distance = distance + d[i]*S_inv_mult_d[i];
			}
			
			return Math.sqrt(distance);
		}
	}

	private double[] doubleMatrixInRMatrix(double[][] S) {
		double[] temp = new double[S.length*S.length];
        int k = 0;
        for(int i = 0; i< S.length; i++){
        	for(int j = 0; j< S.length; j++){
        		temp[k] = S[i][j];
        		k++;
        	}	
        }
        return temp;
	}

	private double[][] calculateFinalCovMatrix(SampleElement[] sampleElements, double a, double[] varVector) {
		int n = sampleElements.length;
		
		double[][] S = new double [n][n];

		for(int i = 0; i<n; i++){
			for(int j = i; j<n; j++){
				double distance = calculateDistance(sampleElements[i], sampleElements[j]);
				S[i][j] = Math.exp(-distance/a)*Math.sqrt(varVector[i]*varVector[j]);
				S[j][i] = S[i][j];
			}
		}
		
		return S;
		
		
		
	}

	private double[] createCovarianceVector(REXP rNormCovMatrix) {
		double[][] covMatrix = rNormCovMatrix.asDoubleMatrix();
		double[] covVector = new double[(covMatrix.length*(covMatrix.length+1))/2];
		int k = 0;
		for(int i = 0; i<covMatrix.length; i++){
			for(int j = i; j<covMatrix.length; j++){
				covVector[k] = covMatrix[i][j];
				k++;
			}
		}
		return covVector;
	}

	private double[] createDistanceVector(List<GeotiffProxy> spatialSet, int[] sampleMask) {
			
		//take one GeotiffProxy to get one sample
		SampleElement[] sample = spatialSet.get(0).getSample();
		
		//n*(n+1)/2 distances to be calculated
		double[] distanceVector = new double[(sampleMask.length*(sampleMask.length+1))/2];
		int k = 0;
		for(int i = 0; i<sampleMask.length; i++){
			for(int j = i; j<sampleMask.length; j++){
				distanceVector[k] = calculateDistance(sample[sampleMask[i]], sample[sampleMask[j]]);
				k++;
			}
		}
		return distanceVector;
	}

	private double calculateDistance(SampleElement se1,
			SampleElement se2) {
		double distance = Math.sqrt(Math.pow(se1.getX()-se2.getX(), 2)+Math.pow(se1.getY()-se2.getY(), 2));
		return distance;
	}

	private int[] createAllSamplesArray(int chNo, List<GeotiffProxy> spatialSet) {

		//no of images in set
		int m = spatialSet.size();
		//no of sample elements used to estimate the regression afterwards
		int n = spatialSet.get(0).getSample().length;
		
		int[] allSamples = new int[m*n];
		
		int iGeotiff = 0;
		
		for(GeotiffProxy geotiff : spatialSet){
			SampleElement[] sample = geotiff.getSample();
			for(int jGeotiff = 0; jGeotiff<n; jGeotiff++){
				//connect all sample-arrays to a big one to use it in R
				allSamples[iGeotiff*n+jGeotiff] = sample[jGeotiff].getValue(chNo);
			}
			
			iGeotiff++;
		}
		
		return allSamples;
	}
	
	private int[] createSamplesArray(int chNo, List<GeotiffProxy> spatialSet, int[] sampleMask) {

		//no of images in set
		int m = spatialSet.size();
		//no of sample elements used to estimate the regression afterwards
		int n = sampleMask.length;
		
		int[] allSamples = new int[m*n];
		
		int iGeotiff = 0;
		
		for(GeotiffProxy geotiff : spatialSet){
			SampleElement[] sample = geotiff.getSample();
			for(int jGeotiff = 0; jGeotiff<n; jGeotiff++){
				//connect all sample-arrays to a big one to use it in R
				allSamples[iGeotiff*n+jGeotiff] = sample[sampleMask[jGeotiff]].getValue(chNo);
			}
			
			iGeotiff++;
		}
		
		return allSamples;
	}
	class TextConsole implements RMainLoopCallbacks
	{
	    public void rWriteConsole(Rengine re, String text, int oType) {
//	        System.out.print(text);
	    }
	     
	    public void rBusy(Rengine re, int which) {
	       // System.out.println("rBusy("+which+")");
	    }
	    
	    public String rReadConsole(Rengine re, String prompt, int addToHistory) {
//	        System.out.print(prompt);
	        try {
	            BufferedReader br=new BufferedReader(new InputStreamReader(System.in));
	            String s=br.readLine();
	            return (s==null||s.length()==0)?s:s+"\n";
	        } catch (Exception e) {
//	            System.out.println("jriReadConsole exception: "+e.getMessage());
	        }
	        return null;
	    }
	    
	    public void rShowMessage(Rengine re, String message) {
//	        System.out.println("rShowMessage \""+message+"\"");
	    }
		
	    public String rChooseFile(Rengine re, int newFile) {
		FileDialog fd = new FileDialog(new Frame(), (newFile==0)?"Select a file":"Select a new file", (newFile==0)?FileDialog.LOAD:FileDialog.SAVE);
		//fd.show();
		String res=null;
		if (fd.getDirectory()!=null) res=fd.getDirectory();
		if (fd.getFile()!=null) res=(res==null)?fd.getFile():(res+fd.getFile());
		return res;
	    }
	    
	    public void   rFlushConsole (Rengine re) {
	    }
		
	    public void   rLoadHistory  (Rengine re, String filename) {
	    }			
	    
	    public void   rSaveHistory  (Rengine re, String filename) {
	    }			
	}
	
	
}
