package de.ifgi.operations;

import java.awt.Color; 
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.imageio.ImageIO;

import de.ifgi.basics.BoundingBox;
import de.ifgi.proxy.GeotiffProxy;
import de.ifgi.proxy.SampleElement;

public class Tiffloader {
	
	private static Tiffloader instance;

    private Tiffloader() {}

    public synchronized static Tiffloader getInstance() {
        if (instance == null) {
            instance = new Tiffloader();
        }
        return instance;
    }
	
	
	public void writeGeotiffs(List<GeotiffProxy> finalSelection){
		for(GeotiffProxy g : finalSelection){
			String adresse = g.getAdresse();
			File source = new File (adresse);
			String filename = adresse.substring(adresse.lastIndexOf("\\"));
			String pathname = adresse.substring(0,adresse.lastIndexOf("\\"));
			String destinationAdresse = pathname+"\\selected"+filename;
			File destination = new File(destinationAdresse); 
			try {
				copyFile(source, destination);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			//Datei aus dem quellOrdner Adresse wird per Default in dessen Unterordner Selected geschoben
		}
	}
	
	public void copyFile(File src, File dest) throws IOException {

	    byte[] buffer = new byte[4096];
	    int read = 0;
	    InputStream in = null;
	    OutputStream out = null;
	    try {
	        in = new FileInputStream(src);
	        out = new FileOutputStream(dest);
	        while(true) {
	            read = in.read(buffer);
	            if (read == -1) {
	                //-1 bedeutet EOF
	                break;
	            }
	            out.write(buffer, 0, read);
	        }
	    } finally {
	        if (in != null) {
	            try {
	                in.close();
	            }
	            finally {
	                if (out != null) {
	                    out.close();
	                }
	            }
	        }
	    }
	}

	public List<GeotiffProxy> loadGeotiffs(double sampleQuota, String[]files) {
		
		List<GeotiffProxy> checkedImages = new ArrayList<GeotiffProxy>();
		
		
		String img = files[0].toString();
		// ist das Buffered Image, aus dem length und width ausgelesen werden
		BufferedImage buffwh;
	
		buffwh = readAsBufferedImage(img);
			
	
		int length = buffwh.getHeight();
		int width = buffwh.getWidth();
		
		int samplesize = new Double(sampleQuota*length*width).intValue();
		//restrict the samplesize
		samplesize = 1000;
		//erstelle eine zufllige Maske, aus der die Samples bestimmt werden
		SampleElement [] sampleMask = createSampleMask(samplesize, length, width);
		
		
		// ld alle Bilder dessen Pfade in files angegeben sind und erstellt daraus ein GeotiffProxy
		for (int i=0; i< files.length; i++){
		
		BufferedImage tempImg = readAsBufferedImage(files[i].toString());	
			
			//TODO: bbox bestimmen
			BoundingBox bbox = null;
			
			//adresse des bildes speichern
			String adresse = files[i];
			//sample mit werten fllen
			SampleElement [] sample = new SampleElement[samplesize];
			
			for(int j = 0; j<samplesize; j++){
				sample[j] = new SampleElement(sampleMask[j].getX(),sampleMask[j].getY());
				
//				TODO: number of channels dynamisch auslesen
//				for(int k = 0; k<channelNumber; k++){
//				}
				//Bildwert an der stelle x,y (Bildkoordinaten) ermitteln
				Color col = new Color(tempImg.getRGB(sampleMask[j].getX(), sampleMask[j].getY()));
				sample[j].setValue(col.getRed(), 0);
				sample[j].setValue(col.getGreen(), 1);
				sample[j].setValue(col.getBlue(), 2);
			}
			GeotiffProxy geotiff = new GeotiffProxy(sample, bbox, adresse);
			checkedImages.add(geotiff);

		}
		return checkedImages;
		
	}
	
	//TODO: erstellt eine Maske von Samples ggf anpassen bei mehreren channels
	private SampleElement[] createSampleMask(int samplesize, int length, int width) {
		SampleElement [] sampleMask = new SampleElement[samplesize];
		Random r = new Random(System.nanoTime());
		for (int i = 0; i<samplesize; i++){
			int x = r.nextInt(width);
			int y = r.nextInt(length);
			sampleMask[i] = new SampleElement(x,y);
		}
		return sampleMask;
	}
	
	public BufferedImage readAsBufferedImage(String img)    {
	    //string to ByteArrayInputStream
	    BufferedImage bImage = null;
	    try {
			FileInputStream fis = new FileInputStream(new File(img));
	  
	        bImage = ImageIO.read(fis);
	    } catch (IOException ex) {
	        Logger.getLogger(Image.class.getName()).log(Level.SEVERE, null, ex);
	    }

	    return bImage;
	}
}
